
#include <stdio.h>

#define NEED_EVENTS
#define NEED_REPLIES
#include <X11/Xlibint.h>
#include <X11/extensions/extutil.h>

/* from "NVCtrlLib.h" */
#define NV_CTRL_VIDEO_RAM 6

/* from "nv_control.h" */
#define NV_CONTROL_NAME "NV-CONTROL"
#define NV_CONTROL_EVENTS 1
typedef struct {
    BYTE type;   /* X_Reply */
    CARD8 padb1;
    CARD16 sequenceNumber B16;
    CARD32 length B32;
    CARD32 isnv B32;
    CARD32 padl4 B32;
    CARD32 padl5 B32;
    CARD32 padl6 B32;
    CARD32 padl7 B32;
    CARD32 padl8 B32;
} xnvCtrlIsNvReply;
//#define sz_xnvCtrlIsNvReply 32
typedef struct {
    CARD8 reqType;
    CARD8 nvReqType;
    CARD16 length B16;
    CARD32 screen B32;
} xnvCtrlIsNvReq;
#define sz_xnvCtrlIsNvReq 8
typedef struct {
    BYTE type;
    BYTE pad0;
    CARD16 sequenceNumber B16;
    CARD32 length B32;
    CARD32 flags B32;
    INT32  value B32;
    CARD32 pad4 B32;
    CARD32 pad5 B32;
    CARD32 pad6 B32;
    CARD32 pad7 B32;
} xnvCtrlQueryAttributeReply;
//#define sz_xnvCtrlQueryAttributeReply 32
typedef struct {
    CARD8 reqType;
    CARD8 nvReqType;
    CARD16 length B16;
    CARD32 screen B32;
    CARD32 display_mask B32;
    CARD32 attribute B32;
} xnvCtrlQueryAttributeReq;
#define sz_xnvCtrlQueryAttributeReq 16
#define X_nvCtrlIsNv                                1
#define X_nvCtrlQueryAttribute                      2

/* from "R200_extensions.h" */
#define ATIFGL_EXTENSION_NAME "ATIFGLEXTENSION"
#define ATIFGL_NUMBER_EVENTS 0
typedef struct _FGLGetDriverData {
    CARD8   reqType;            //    1
    CARD8   fireglReqType;      //    1
    CARD16  length B16;         //    2
    /* specific */
    CARD32  screen B32;         //    4
    CARD16  size B16;           //    2
    CARD16  pad1;               //    2
} xFGLGetDriverDataReq;         // = 12
#define sz_xFGLGetDriverDataReq      12/*XREQ_SZ(FGLGetDriverData)*/
typedef struct {
    BYTE    type;               //    1    (X_Reply)
    BYTE    pad1;               //    1
    CARD16  sequenceNumber B16; //    2
    CARD32  length B32;         //    4
    /* specific */
    CARD8   majorVersion;       //    1
    CARD8   minorVersion;       //    1
    CARD8   patchlevel B16;     //    1
    CARD8   BIOSVersionMajor;   //    1
    CARD8   BIOSVersionMinor;   //    1
    CARD8   HasSecondary;       //    1
    CARD16  pad3 B16;           //    2
    CARD16  usBoardType B16;    //    2
    CARD16  usChipType  B16;    //    2
    CARD16  usVideoRam  B16;    //    2
    CARD8   sATiRevID[17];      //   17 (previousely 1, now 1+16 to fullfill alignment)
    CARD8   AGPTransferMode;    //    1
    CARD32  AGPCapPtr;          //    4
    CARD32  AGPStatus;          //    4
    CARD32  AGPCommand;         //    4
    CARD32  ulGamma1  B32;      //    4
    CARD32  ulGamma2  B32;      //    4
    CARD32  ulDriverType  B32;  //    4
    CARD32  ulDesktopSetup B32; //    4
    CARD32  ulPrimary B32;      //    4
} xFGLGetDriverDataReply;       // = 72
// 72 bytes 
#define X_FGLGetDriverData 0

static XExtensionInfo _ctrl_ext_info_data;
static XExtensionInfo *ctrl_ext_info = &_ctrl_ext_info_data;
static /* const */ XExtensionHooks ctrl_extension_hooks = {
    NULL,                               /* create_gc */
    NULL,                               /* copy_gc */
    NULL,                               /* flush_gc */
    NULL,                               /* free_gc */
    NULL,                               /* create_font */
    NULL,                               /* free_font */
    NULL,                               /* close_display */
    NULL,                               /* wire_to_event */
    NULL,                               /* event_to_wire */
    NULL,                               /* error */
    NULL,                               /* error_string */
};

int getVideoRam(Display *dpy)
{
	int vram;
	XExtDisplayInfo *info;

	/* Try the NV-CONTROL x-server extension: */
	info = XextAddDisplay (ctrl_ext_info, dpy, NV_CONTROL_NAME, &ctrl_extension_hooks, NV_CONTROL_EVENTS, NULL);
	if(XextHasExtension(info))
	{
		xnvCtrlIsNvReply rep1;
		xnvCtrlIsNvReq   *req1;
		xnvCtrlQueryAttributeReply rep2;
		xnvCtrlQueryAttributeReq   *req2;
		
		LockDisplay (dpy);
		GetReq (nvCtrlIsNv, req1);
		req1->reqType = info->codes->major_opcode;
		req1->nvReqType = X_nvCtrlIsNv;
		req1->screen = DefaultScreen(dpy);
		if (_XReply (dpy, (xReply *) &rep1, 0, xTrue))
		{
			if (rep1.isnv)
			{
				GetReq (nvCtrlQueryAttribute, req2);
				req2->reqType = info->codes->major_opcode;
				req2->nvReqType = X_nvCtrlQueryAttribute;
				req2->screen = req1->screen;
				req2->display_mask = 0;
				req2->attribute = NV_CTRL_VIDEO_RAM;
				if (_XReply (dpy, (xReply *) &rep2, 0, xTrue))
				{
					if (rep2.flags)
					{
						UnlockDisplay (dpy);
						SyncHandle ();
						XextRemoveDisplay(ctrl_ext_info, dpy);
						return rep2.value;
					}
				}
			}
		}
		UnlockDisplay (dpy);
		SyncHandle ();
	}
	XextRemoveDisplay(ctrl_ext_info, dpy);

	/* Try the ATIFGLEXTENSION x-server extension: */
	info = XextAddDisplay (ctrl_ext_info, dpy, ATIFGL_EXTENSION_NAME, &ctrl_extension_hooks, ATIFGL_NUMBER_EVENTS, NULL);
	if(XextHasExtension(info))
	{
		xFGLGetDriverDataReq *req;
		xFGLGetDriverDataReply rep;
		int i;

		LockDisplay (dpy);
		GetReq(FGLGetDriverData, req);

		// Fill in the request structure
		req->reqType = info->codes->major_opcode;
		// calculate size of request structure in 4 byte units 
		req->length = (sizeof(*req)+3) / 4;
		req->fireglReqType = X_FGLGetDriverData;
		req->screen = DefaultScreen(dpy);

		// Carry out the extension function and wait for reply
		// 3rd parameter, called "extra": number of extra 32 bit words
		// reply structures must be multiple of 4 with a minimum of 32 bytes
		if (_XReply(dpy, (xReply *)&rep, (sizeof(rep) - 32) / 4, xFalse))
		{
			UnlockDisplay (dpy);
			SyncHandle();
			XextRemoveDisplay(ctrl_ext_info, dpy);
			return rep.usVideoRam;
		}
		UnlockDisplay (dpy);
		SyncHandle();
	}
	XextRemoveDisplay(ctrl_ext_info, dpy);

	return 0;
}

int main(int argc, char **argv)
{
	Display *dpy=XOpenDisplay(0);
	printf("videoRam: %i kBytes\n", getVideoRam(dpy));
	XCloseDisplay(dpy);
}

